;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "doom-modeline" "20260411.1454"
  "A minimal and modern mode-line."
  '((emacs       "25.1")
    (compat      "30.1.0.0")
    (nerd-icons  "0.1.0")
    (shrink-path "0.3.1"))
  :url "https://github.com/seagle0128/doom-modeline"
  :commit "870b4b215fda881d29278d81bbab7f934a9bd8b3"
  :revdesc "870b4b215fda"
  :keywords '("faces" "mode-line")
  :authors '(("Vincent Zhang" . "seagle0128@gmail.com"))
  :maintainers '(("Vincent Zhang" . "seagle0128@gmail.com")))
